package org.expeditee.gio.input;

import org.expeditee.actions.Actions;
import org.expeditee.actions.Simple;
import org.expeditee.gio.InputManager;
import org.expeditee.gio.input.InputEvent.InputType;
import org.expeditee.gio.input.KBMInputEvent.EventType;
import org.expeditee.gio.input.KBMInputEvent.Key;
import org.expeditee.gio.input.KBMInputEvent.MouseButton;
import org.expeditee.gui.FrameUtils;
import org.expeditee.gui.Help;
import org.expeditee.stats.Logger;
import org.expeditee.stats.SessionStats;

/**
 * Static class to house all of the standard Expeditee input listeners.
 * 
 * @author cts16
 */
public class StandardInputEventListeners {

	/** Cannot be constructed, static class only. */
	private StandardInputEventListeners()
	{
	}
	
	/** Registers the standard input listeners with the given input manager. */
	public static void registerWith(InputManager manager)
	{
		if (manager == null) return;
		
		// Register the listeners in priority order
		manager.registerInputEventListener(kbmStateListener);
		manager.registerInputEventListener(helpListener);
		manager.registerInputEventListener(statisticsListener);
		manager.registerInputEventListener(agentListener);
		manager.registerInputEventListener(simpleListener);
	}
	
	/*
	 * 
	 * Expeditee's standard input-event listeners.
	 * 
	 */
	
	/** An input listener to track the state of all keys and mouse buttons. */
	public static final KBMStateListener kbmStateListener = new KBMStateListener();
	public static class KBMStateListener implements InputEventListener
	{
		private Boolean[] _keyState = new Boolean[Key.values().length];
		private Boolean[] _mouseState = new Boolean[MouseButton.values().length];
		
		@Override
		public boolean onInputEvent(InputEvent event) {
			// Only interested in keyboard/mouse events
			if (event.getInputType() != InputType.KBM) return false;
			
			// Cast the event
			KBMInputEvent kbmEvent = (KBMInputEvent) event;
			
			// Track any key state changes
			if (kbmEvent.getKeyDown() != null) {
				_keyState[kbmEvent.getKeyDown().ordinal()] = true;
			} else if (kbmEvent.getKeyUp() != null) {
				_keyState[kbmEvent.getKeyUp().ordinal()] = false;
			}
			
			// Track any mouse button state changes
			if (kbmEvent.getMouseButtonDown() != null) {
				_mouseState[kbmEvent.getMouseButtonDown().ordinal()] = true;
			} else if (kbmEvent.getMouseButtonUp() != null) {
				_mouseState[kbmEvent.getMouseButtonUp().ordinal()] = false;
			}
			
			return false;
		}
		
		public boolean isKeyDown(Key key) {
			return (key != null) && (_keyState[key.ordinal()] == Boolean.TRUE);
		}
		
		public boolean isMouseButtonDown(MouseButton button) {
			return (button != null) && (_mouseState[button.ordinal()] == Boolean.TRUE);
		}
		
		public boolean isOnlyMouseButtonDown(MouseButton button) {
			if (button == null) return false;
			
			for (MouseButton b : MouseButton.values()) {
				if ((b == button && !isMouseButtonDown(b)) || (b != button && isMouseButtonDown(b))) return false;
			}
			
			return true;
		}
		
		public boolean isAnyMouseButtonDown() {
			for (MouseButton button : MouseButton.values()) {
				if (isMouseButtonDown(button)) return true;
			}
			
			return false;
		}
	};
	
	/** An input listener to monitor the input for statistical purposes. */ 
	private static InputEventListener statisticsListener = new InputEventListener()
	{
		@Override
		public boolean onInputEvent(InputEvent event) {
			// Only interested in keyboard/mouse events
			if (event.getInputType() != InputEvent.InputType.KBM) return false;
			KBMInputEvent kbmEvent = (KBMInputEvent) event;
			if (kbmEvent.getEventType() == EventType.KEY_DOWN) {
				// Notify the change in stats
				// TODO: Will changing this to the Expeditee KBMInputEvent upset stats? cts16
				SessionStats.AddFrameEvent("k" + kbmEvent.getKeyText());

				// Used for calculating frame stats (response time)
				FrameUtils.ResponseTimer.restart();
				
				// Watch for escape key
				if (kbmEvent.getKeyDown() == Key.ESC) SessionStats.Escape();
				
				if (kbmStateListener.isKeyDown(Key.CTRL)) {
					Logger.Log(Logger.USER, Logger.CONTROL_CHAR, "User pressing: Ctrl+" + kbmEvent.getKeyText());
				}
			} else if (kbmEvent.getEventType() == EventType.MOUSE_BUTTON_DOWN) {
				SessionStats.MouseClicked(kbmEvent.getMouseButtonDown());
				switch (kbmEvent.getMouseButtonDown()) {
				case LEFT:
					SessionStats.AddFrameEvent("Ld");
					break;
				case MIDDLE:
					SessionStats.AddFrameEvent("Md");
					break;
				case RIGHT:
					SessionStats.AddFrameEvent("Rd");
					break;
				default:
					break;
				}
			} else if (kbmEvent.getEventType() == EventType.MOUSE_BUTTON_UP) {
				FrameUtils.ResponseTimer.restart();
				switch (kbmEvent.getMouseButtonUp()) {
				case LEFT:
					SessionStats.AddFrameEvent("Lu");
					break;
				case MIDDLE:
					SessionStats.AddFrameEvent("Mu");
					break;
				case RIGHT:
					SessionStats.AddFrameEvent("Ru");
					break;
				default:
					break;
				}
			}
			return false;
		}
	};
				
	/** An input listener to monitor keyboard input for running agents. */
	private static InputEventListener agentListener = new InputEventListener() {
		@Override
		public boolean onInputEvent(InputEvent event) {
			// Only interested in keyboard/mouse events
			if (event.getInputType() != InputType.KBM) return false;
			KBMInputEvent kbmEvent = (KBMInputEvent) event;
			if (Actions.isAgentRunning()) {
				if (kbmEvent.getKeyDown() == Key.ESC || kbmEvent.getEventType() == EventType.MOUSE_BUTTON_DOWN)	{
					Actions.stopAgent();
				} else if (kbmEvent.getEventType() == EventType.KEY_DOWN) {
					Actions.interruptAgent();
				}
			}
			return false;
		}
	};
				
	/** An input listener to consume keyboard input for running Simple programs. */
	private static InputEventListener simpleListener = new InputEventListener() {
		@Override
		public boolean onInputEvent(InputEvent event) {
			// Only interested in keyboard/mouse events
			if (event.getInputType() != InputEvent.InputType.KBM) return false;
			if (Simple.consumeKeyboardInput()) return true;
			return false;
		}
	};
	
	/** An input listener to update the message-bay help text. */
	private static InputEventListener helpListener = new InputEventListener()
	{
		@Override
		public boolean onInputEvent(InputEvent event) {
			// Only interested in keyboard/mouse events
			if (event.getInputType() != InputEvent.InputType.KBM) return false;
			Help.updateStatus(kbmStateListener.isKeyDown(Key.CTRL), kbmStateListener.isKeyDown(Key.SHIFT));
			return false;
		}
	};
}
